/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.specific;

import java.io.File;
import java.io.StringReader;
import java.util.Enumeration;

import org.w3c.dom.DocumentType;

import ch.cern.lhcb.xmleditor.Editor;
import ch.cern.lhcb.xmleditor.dtdparser.CustomDTDParser;
import ch.cern.lhcb.xmleditor.gui.MessageDisplayer;

import com.wutka.dtd.DTD;

/**
 * This class is a kind of accessor for XmlDisplays. It mainly implements a
 * cache on XmlDisplays.
 * 
 * @author Sebastien Ponce
 */

public class XmlDisplaysManager {

    /**
     * this pool keeps every XmlDisplay opened together with the number of
     * places where it is used.
     */
    private Pool refTable;

    /**
     * the editor using this Files Manager
     */
    private Editor xmlEditor;

    /**
     * Default constructor
     * 
     * @param xmlEditor the XmlEditor using this XmlDisplaysManager
     */
    public XmlDisplaysManager(Editor xmlEditor) {
        super();
        this.xmlEditor = xmlEditor;
        refTable = new Pool();
    }

    /**
     * gets a XmlDisplay for an Xml file It first gets the dtd, then creates an
     * XmlDisplay
     * 
     * @param xmlFile the name of the xml file we want to get the associated
     *        xmlDisplay
     * @param msgDis the associated MessageDisplayer (for error displaying)
     * @return the XmlDisplay or null if no dtd was available
     */
    public XmlDisplay getXmlDisplay(XmlFile xmlFile, MessageDisplayer msgDis) {
        DocumentType documentType = xmlFile.getDocument().getDoctype();
        if (documentType == null) {
            msgDis.popupMessage("Invalid document");
            msgDis.displayMessage("The file " + xmlFile
                    + " does not define a valid Xml document. "
                    + "You'd better not edit it here.\n");
            return null;
        }
        // if an internal Dtd exists, parses it
        DTD internalDtd = parseDtd(documentType.getInternalSubset(), xmlFile,
                msgDis);
        // Deals with the external subset of the Dtd : gets the fileName of the
        // external subset of the Dtd
        String fileName = documentType.getSystemId();
        // check whether we have an absolute file name or not
        File tempFile = new File(fileName);
        boolean isAbsoluteFileName = tempFile.isAbsolute();
        // gets the dtd file
        DtdFile dtdFile;
        if (isAbsoluteFileName) {
            dtdFile = new DtdFile(fileName, msgDis);
        } else {
            dtdFile = new DtdFile(xmlFile, fileName, msgDis);
        }
        // look whether we can reuse an existing XmlDisplay
        boolean reuseOk = false;
        XmlDisplay display = null;
        Enumeration enm = refTable.getPoolContent();
        while (enm.hasMoreElements()) {
            XmlDisplay existingDisplay = (XmlDisplay) enm.nextElement();
            boolean dtdFileEqual = false;
            DtdFile existingDtdFile = existingDisplay.getDtdFile();
            if (existingDtdFile == null) {
                dtdFileEqual = dtdFile == null;
            } else {
                dtdFileEqual = existingDtdFile.equals(dtdFile);
            }
            boolean internalDtdEqual = false;
            DTD existingInternalDtd = existingDisplay.getInternalDtd();
            if (existingInternalDtd == null) {
                internalDtdEqual = internalDtd == null;
            } else {
                internalDtdEqual = existingInternalDtd.equals(internalDtd);
            }
            if (dtdFileEqual && internalDtdEqual) {
                display = existingDisplay;
                reuseOk = true;
                break;
            }
        }
        // if we cannot reuse, builds a new XmlDisplay
        if (!reuseOk) {
            if (isAbsoluteFileName) {
                dtdFile = xmlEditor.getDtdFilesManager().getDtdFile(fileName,
                        msgDis);
            } else {
                dtdFile = xmlEditor.getDtdFilesManager().getDtdFile(xmlFile,
                        fileName, msgDis);
            }
            display = new XmlDisplay(dtdFile, internalDtd);
        }
        // adds a reference on the new XmlDisplay
        refTable.addRef(display);
        // if no Dtd at all, let the user know that the functionnalities will
        // be reduced
        if (display.getDtdFile() == null && display.getInternalDtd() == null) {
            msgDis.popupMessage("No Dtd available");
            msgDis
                    .displayMessage("Since no Dtd is available, the functionnalities will be reduced.\n"
                            + "For example, you should not be able to add nodes or attributes.\n");
            return null;
        }
        return display;
    }

    /**
     * parses the internal dtd if it exists from a string
     * 
     * @param dtd the string containing the dtd to parse
     * @param file the file containing this dtd
     * @param msgDis the associated MessageDisplayer (for error displaying)
     * @return the parsed DTD or null if parsing was not successfull
     */
    private DTD parseDtd(String dtd, XmlFile file, MessageDisplayer msgDis) {
        if (dtd == null) { return null; }
        try {
            CustomDTDParser parser = new CustomDTDParser(new StringReader(dtd),
                    file.getParentFile());
            return parser.parse();
        } catch (java.io.IOException e) {
            msgDis.popupMessage("Error while parsing internal Dtd");
            msgDis
                    .displayMessage("An error occured while parsing internal Dtd :\n"
                            + e + "\n");
        }
        return null;
    }

    /**
     * called when an xmlDisplay is closed. The XmlDisplay will only be closed
     * if nobody else uses it.
     * 
     * @param display the XmlDisplay to close
     */
    public void closeDisplay(XmlDisplay display) {
        if (display == null) { return; }
        refTable.removeRef(display);
        if (refTable.getRefNumber(display) == 0) {
            refTable.remove(display);
            xmlEditor.getDtdFilesManager().closeFile(display.getDtdFile());
        }
    }

}