/*
 * com.simulogic.dataview -- The Java2D Plotting Framework
 *
 * Copyright (C) 1999 - 2005 Sergey Missan
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.simulogic.dataview;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Vector;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;
import javax.swing.ToolTipManager;

import com.simulogic.dataview.actions.AxisFullScaleAction;
import com.simulogic.dataview.actions.AxisPropertiesAction;
import com.simulogic.dataview.actions.AxisZoomInAction;
import com.simulogic.dataview.actions.AxisZoomOutAction;

/**
 * This class displays data axis for the given <code>DataView</code>
 */
public class DataAxis extends javax.swing.JPanel implements
		javax.swing.event.ChangeListener, java.awt.event.MouseListener,
		java.awt.event.MouseMotionListener {

	/**
	 * Specifies horizontal orientation of axis
	 */
	public static final int AXIS_HORIZONTAL = 0;

	/**
	 * Specifies vertical orientation of axis
	 */
	public static final int AXIS_VERTICAL = 1;

	private int orientation;

	private DataView view;

	private JPopupMenu popup;

	private Rectangle2D.Double zoomBounds = new Rectangle2D.Double();

	private Graphics2D zoomGraphics;

	private Vector ticks;

	private Vector labels;

	private int labelUnitWidth;

	private int labelUnitHeight;

	private int digits;

	private boolean zeroCross;

	private boolean showGrid = true;

	private double tickLength;

	private Color scaleColor;

	private Color baseColor;

	private Color ticksColor;

	private java.text.DecimalFormat format;

	private java.awt.geom.AffineTransform transform = new java.awt.geom.AffineTransform();

	/**
	 * Creates a new axis
	 * 
	 * @param o
	 *            axis orientation
	 * @param v
	 *            data view that hosts this axis
	 */
	public DataAxis(int o, DataView v) {
		super();

		orientation = o;
		view = v;

		this.addMouseListener(this);
		this.addMouseMotionListener(this);

		if (orientation == AXIS_HORIZONTAL) {
			view.getScrollPane().setColumnHeaderView(this);
			view.getScrollPane().getColumnHeader().addChangeListener(this);
		} else {
			view.getScrollPane().setRowHeaderView(this);
			view.getScrollPane().getRowHeader().addChangeListener(this);
		}

		ticks = new Vector();
		labels = new Vector();
		tickLength = 5.0;

		format = new java.text.DecimalFormat("#.###");
		format.setMaximumFractionDigits(3);

		setBaseColor(getBackground());
		setTicksColor(getForeground());
		setScaleColor(getForeground().brighter().brighter());

		ToolTipManager.sharedInstance().registerComponent(this);
	}

	/**
	 * Returns hosting view
	 * 
	 * @return view
	 */
	public DataView getView() {
		return view;
	}

	/**
	 * Returns axis orientation
	 * 
	 * @return axis orientation
	 */
	public int getOrientation() {
		return orientation;
	}

	/**
	 * Returns tooltip with information about axis
	 * 
	 * @return tooltip string
	 */
	public String getToolTipText() {
		StringBuffer tip = new StringBuffer();
		if (this.orientation == AXIS_HORIZONTAL) {
			tip.append("<X> ");
			tip.append(format.format(view.getScale().getX()));
			tip.append(" : ");
			tip.append(format.format(view.getScale().getX()
					+ view.getScale().getWidth()));
			tip.append(" (" + view.getXCorner().getText() + ")");
		} else {
			tip.append("<Y> ");
			tip.append(format.format(view.getScale().getY()
					- view.getScale().getHeight()));
			tip.append(" : ");
			tip.append(format.format(view.getScale().getY()));
			tip.append(" (" + view.getYCorner().getText() + ")");
		}
		return tip.toString();
	}

	/**
	 * Sets new font
	 * 
	 * @param f
	 *            new font
	 */
	public void setFont(java.awt.Font f) {
		super.setFont(f);

		if (view != null)
			changeLimits();
	}

	/**
	 * Sets axis to zero cross
	 * 
	 * @param zeroCross
	 *            new value of zero cross
	 */
	public void setZeroCross(boolean zeroCross) {
		this.zeroCross = zeroCross;
	}

	/**
	 * Returns if axis is zero cross
	 * 
	 * @return zero cross
	 */
	public boolean isZeroCross() {
		return zeroCross;
	}

	/**
	 * Returns tick length
	 * 
	 * @return tick length
	 */
	public double getTickLength() {
		return tickLength;
	}

	/**
	 * Sets tick length
	 * 
	 * @param tickLength
	 *            new value of tick length
	 */
	public void setTickLength(int tickLength) {
		this.tickLength = tickLength;
	}

	/**
	 * Getter for property scaleColor.
	 * 
	 * @return Value of property scaleColor.
	 */
	public java.awt.Color getScaleColor() {
		return scaleColor;
	}

	/**
	 * Setter for property scaleColor.
	 * 
	 * @param scaleColor
	 *            New value of property scaleColor.
	 */
	public void setScaleColor(java.awt.Color scaleColor) {
		this.scaleColor = scaleColor;
	}

	/**
	 * Getter for property baseColor.
	 * 
	 * @return Value of property baseColor.
	 */
	public java.awt.Color getBaseColor() {
		return baseColor;
	}

	/**
	 * Setter for property baseColor.
	 * 
	 * @param baseColor
	 *            New value of property baseColor.
	 */
	public void setBaseColor(java.awt.Color baseColor) {
		this.baseColor = baseColor;
	}

	/**
	 * Getter for property ticksColor.
	 * 
	 * @return Value of property ticksColor.
	 */
	public java.awt.Color getTicksColor() {
		return ticksColor;
	}

	/**
	 * Setter for property ticksColor.
	 * 
	 * @param ticksColor
	 *            New value of property ticksColor.
	 */
	public void setTicksColor(java.awt.Color ticksColor) {
		this.ticksColor = ticksColor;
	}

	/**
	 * Returns current ticks positions
	 * 
	 * @return vector of current ticks
	 */
	public java.util.Vector getTicks() {
		return ticks;
	}

	/**
	 * Setter for property showGrid.
	 * 
	 * @param showGrid
	 *            New value of property showYGrid.
	 */
	public void setShowGrid(boolean showGrid) {
		this.showGrid = showGrid;
	}

	/**
	 * Getter for property showGrid.
	 * 
	 * @return Value of property showGrid.
	 */
	public boolean isShowGrid() {
		return showGrid;
	}

	/**
	 * Changes the scale for the axis
	 * 
	 * @param rsc
	 *            scale rectangle
	 */
	public void setScale(Rectangle2D.Double rsc) {
		if (orientation == AXIS_HORIZONTAL) {
			view.setScale(rsc.x, view.getScale().getY(), rsc.width, view
					.getScale().getHeight());
		} else {
			view.setScale(view.getScale().getX(), rsc.y, view.getScale()
					.getWidth(),

			rsc.height);
		}
	}

	/**
	 * Rescales the axis to full scale
	 */
	public void fullScale() {
		if (orientation == AXIS_HORIZONTAL) {
			view.setScale(view.getStorage().getLimitRectangle().getX(), view
					.getScale().getY(), view.getStorage().getLimitRectangle()
					.getWidth(), view.getScale().getHeight());
		} else {
			view.setScale(view.getScale().getX(), view.getStorage()
					.getLimitRectangle().getY(), view.getScale().getWidth(),
					view.getStorage().getLimitRectangle().getHeight());
		}
	}

	/**
	 * Zooms around center by a certain factor
	 * 
	 * @param factor
	 *            zoom factor. If positive, zooms in; if negative, zooms out
	 */
	public void zoomCentered(double factor) {
		double mult_f = factor;

		if (factor >= 0) {
			mult_f = 1.0 / factor;
		}

		if (orientation == AXIS_HORIZONTAL) {
			view
					.setScale(
							(view.getScale().getX() + (view.getScale()
									.getWidth() / 2.0))
									- Math
											.abs((view.getScale().getWidth() * mult_f) / 2.0),
							view.getScale().getY(), view.getScale().getWidth()
									* mult_f, view.getScale().getHeight());
		} else {
			view
					.setScale(
							view.getScale().getX(),
							(view.getScale().getY() - (view.getScale()
									.getHeight() / 2.0))
									+ Math
											.abs((view.getScale().getHeight() * mult_f) / 2.0),
							view.getScale().getWidth(), view.getScale()
									.getHeight()
									* mult_f);

		}
	}

	/**
	 * Creates and displays dialog for changing axis propertis
	 */
	public void showPropertiesDialog() {
		new com.simulogic.dataview.AxisProperties(null, this).setVisible(true);
	}

	/**
	 * Overrides correspondent JComponent method
	 * 
	 * @param g
	 *            axis graphics
	 */
	public void paintComponent(java.awt.Graphics g) {
		super.paintComponent(g);

		Graphics2D g2 = (Graphics2D) g;

		if (!g.getClipBounds().isEmpty()) {
			if (orientation == AXIS_HORIZONTAL) {
				plotHorizontalAxis(g2);
			} else {
				plotVerticalAxis(g2);
			}
		}
	}

	private boolean changeTransform() {
		if (orientation == AXIS_HORIZONTAL) {
			double ddx = view.getScrollPane().getViewport().getExtentSize()
					.getWidth()
					/ view.getScale().getWidth();
			if (!Double.isInfinite(ddx) && !Double.isNaN(ddx)) {

				if (transform.getScaleX() != ddx) {
					transform.setToScale(ddx, 1.0);
					transform.translate(-Math
							.min(view.getStorage().getLimitRectangle().getX(),
									view.getScale().getX()), 0.0);

					return true;
				}
			}
		} else {
			double ddy = view.getScrollPane().getViewport().getExtentSize()
					.getHeight()
					/ view.getScale().getHeight();
			if (!Double.isInfinite(ddy) && !Double.isNaN(ddy)) {

				if (transform.getScaleY() != -ddy) {
					transform.setToScale(1.0, -ddy);
					transform.translate(0.0, -Math
							.max(view.getStorage().getLimitRectangle().getY(),
									view.getScale().getY()));

					return true;
				}
			}
		}
		return false;
	}

	private void changeLimits() {
		try {
			if (view.getStorage() == null)
				return;

			changeTransform();

			if (orientation == AXIS_HORIZONTAL) {
				int d = createTicks(view.getStorage().getLimitRectangle()
						.getX(), view.getStorage().getLimitRectangle().getX()
						+ view.getStorage().getLimitRectangle().getWidth(),
						view.getScale().getX(), view.getScale().getX()
								+ view.getScale().getWidth(), view.getScale()
								.getWidth());

				setPreferredWidth(d);

				this.repaint();
			} else {
				int d = createTicks(view.getStorage().getLimitRectangle()
						.getY()
						- view.getStorage().getLimitRectangle().getHeight(),
						view.getStorage().getLimitRectangle().getY(), view
								.getScale().getY()
								- view.getScale().getHeight(), view.getScale()
								.getY(), view.getScale().getHeight());

				setPreferredHeight(d);

				this.repaint();
			}
		} catch (Exception e) {
		}
	}

	private void setPreferredWidth(int d) {
		int lw = getFontMetrics(this.getFont()).charWidth('8');
		int lh = getFontMetrics(this.getFont()).getHeight();

		boolean layout = false;

		int hl = view.getPreferredSize().width;

		if ((lw != labelUnitWidth) || (lh != labelUnitHeight)) {
			labelUnitWidth = lw;
			labelUnitHeight = lh;
			layout = true;
		}

		if (d != digits) {
			digits = d;
			layout = true;
		}

		if ((hl != getPreferredSize().width) || layout) {
			setPreferredSize(new Dimension(hl,
					(int) (labelUnitHeight + tickLength * 2)));
			this.revalidate();
		}
	}

	private void setPreferredHeight(int d) {
		int lw = getFontMetrics(this.getFont()).charWidth('8');
		int lh = getFontMetrics(this.getFont()).getHeight();

		boolean layout = false;

		int vl = view.getPreferredSize().height;

		if ((lw != labelUnitWidth) || (lh != labelUnitHeight)) {
			labelUnitWidth = lw;
			labelUnitHeight = lh;
			layout = true;
		}

		if (d != digits) {
			digits = d;
			layout = true;
		}

		if ((vl != getPreferredSize().height) || layout) {
			setPreferredSize(new Dimension((int) (labelUnitWidth * digits
					+ tickLength * 2 + 2), vl));
			this.revalidate();
		}
	}

	private int createTicks(double minlim, double maxlim, double min,
			double max, double size) {

		if (!this.isVisible())
			return 0;

		int digits = 0;
		String label = "";

		ticks.clear();
		labels.clear();
		double step = calculateStep(min, max, (max - min) / size);
		double inc = step * Math.ceil(min / step) - step;

		label = format.format(minlim).trim();
		digits = Math.max(digits, label.length());
		label = format.format(maxlim).trim();
		digits = Math.max(digits, label.length());

		if (min >= minlim) {
			ticks.addElement(new Double(minlim));
			label = format.format(minlim).trim();
			labels.addElement(label);
		}

		while (inc < (max + step * 2.0)) {
			ticks.addElement(new Double(inc));
			label = format.format(inc).trim();
			labels.addElement(label);
			digits = Math.max(digits, label.length());
			inc += step;
		}

		if (max <= maxlim) {
			ticks.addElement(new Double(maxlim));
			label = format.format(maxlim).trim();
			labels.addElement(label);
		}

		return digits;
	}

	private double calculateStep(double min, double max, double adj) {
		double tic;
		double range = log10(Math.abs(max - min));
		double fl = Math.floor(range);
		double xnorm = Math.pow(20.0, range - fl);
		double posns = 20.0 * adj / xnorm;

		if (posns > 100)
			tic = 0.001; // 0, .001, .002, ...
		else if (posns > 40)
			tic = 0.05; // 0, .05, .10, ...
		else if (posns > 20)
			tic = 0.1; // 0, .1, .2, ...
		else if (posns > 10)
			tic = 0.2; // 0, .2, .4, ...
		else if (posns > 4)
			tic = 0.5; // 0, .5, 1, ...
		else if (posns > 1)
			tic = 1; // 0, 1, 2, ...
		else if (posns > 0.5)
			tic = 2; // 0, 2, 4, 6, ...
		else
			tic = Math.ceil(xnorm);

		tic *= Math.pow(10.0, fl);

		return tic;
	}

	private double log10(double x) {
		return Math.log(x) / Math.log(10.0);
	}

	private void plotHorizontalAxis(Graphics2D g2) {
		double yy, xp = 0.0;
		double val;
		String dat;
		java.awt.Rectangle drawHere = g2.getClipBounds();
		Point2D.Double dst = new Point2D.Double();
		GeneralPath gp = new GeneralPath();

		try {
			yy = getSize().height;
			g2.setStroke(new java.awt.BasicStroke(1.0f));
			g2.setFont(this.getFont());

			g2.setColor(getBaseColor());
			g2.drawLine(drawHere.x, (int) yy - 1, drawHere.x + drawHere.width,
					(int) yy - 1);

			g2.setColor(getForeground());
			for (int i = 0; i < ticks.size(); i++) {
				val = ((Double) ticks.get(i)).doubleValue();
				dat = (String) labels.get(i);

				if (i == 0) {
					transform.transform(new Point2D.Double(val, yy - tickLength
							* 2), dst);
					gp.moveTo((float) val, (float) (yy - tickLength));
					gp.lineTo((float) val, (float) yy);

					g2.drawString(dat, (float) dst.getX(), (float) dst.getY());
					xp = dst.getX() + labelUnitWidth * dat.length();
				} else {
					transform.transform(new Point2D.Double(val, yy - tickLength
							* 2), dst);
					if (dst.getX() - labelUnitWidth * dat.length() > xp) {
						gp.moveTo((float) val, (float) (yy - tickLength));
						gp.lineTo((float) val, (float) yy);

						g2.drawString(dat, (float) dst.getX() - labelUnitWidth
								* dat.length(), (float) dst.getY());
						xp = dst.getX();
					}
				}
			}
			g2.setColor(getTicksColor());
			g2.draw(gp.createTransformedShape(transform));

			if (zeroCross) {
				gp.reset();
				g2.setColor(getTicksColor());
				transform.transform(new Point2D.Double(0.0, yy), dst);
				gp.moveTo((float) dst.getX(), (float) yy);
				gp.lineTo((float) (dst.getX() + tickLength),
						(float) (yy - tickLength));
				gp.lineTo((float) (dst.getX() - tickLength),
						(float) (yy - tickLength));
				gp.lineTo((float) dst.getX(), (float) yy);

				g2.fill(gp);
			}

		} catch (Exception e) {
		}
	}

	private void plotVerticalAxis(Graphics2D g2) {
		double xx, yp = 0.0;
		double val;
		String dat;
		java.awt.Rectangle drawHere = g2.getClipBounds();
		Point2D.Double dst = new Point2D.Double();
		GeneralPath gp = new GeneralPath();

		try {
			xx = getSize().width;
			g2.setStroke(new java.awt.BasicStroke(1.0f));
			g2.setFont(this.getFont());

			g2.setColor(getBaseColor());
			g2.drawLine((int) xx - 1, drawHere.y, (int) xx - 1, drawHere.y
					+ drawHere.height);

			g2.setColor(getForeground());
			for (int i = (ticks.size() - 1); i >= 0; i--) {
				val = ((Double) ticks.get(i)).doubleValue();
				dat = (String) labels.get(i);

				if (i == (ticks.size() - 1)) {
					transform.transform(new Point2D.Double(xx - labelUnitWidth
							* dat.length() - tickLength * 2, val), dst);
					gp.moveTo((float) (xx - tickLength), (float) val);
					gp.lineTo((float) xx, (float) val);

					g2.drawString(dat, (float) dst.getX(), (float) dst.getY()
							+ labelUnitHeight - 1);
					yp = dst.getY() + labelUnitHeight - 1;
				} else {
					transform.transform(new Point2D.Double(xx - labelUnitWidth
							* dat.length() - tickLength * 2, val), dst);
					if (dst.getY() - labelUnitHeight > yp) {
						gp.moveTo((float) (xx - tickLength), (float) val);
						gp.lineTo((float) xx, (float) val);

						g2.drawString(dat, (float) dst.getX(), (float) dst
								.getY() - 1);
						yp = dst.getY() - 1;
					}
				}
			}
			g2.setColor(getTicksColor());
			g2.draw(gp.createTransformedShape(transform));

			if (zeroCross) {
				gp.reset();
				g2.setColor(getTicksColor());
				transform.transform(new Point2D.Double(xx, 0.0), dst);
				gp.moveTo((float) xx, (float) dst.getY());
				gp.lineTo((float) (xx - tickLength),
						(float) (dst.getY() + tickLength));
				gp.lineTo((float) (xx - tickLength),
						(float) (dst.getY() - tickLength));
				gp.lineTo((float) xx, (float) dst.getY());

				g2.fill(gp);
			}
		} catch (Exception e) {
		}
	}

	private void flipFrame(Rectangle2D.Double b) {
		drawFrame(zoomBounds);
		zoomBounds.setRect(b);
		drawFrame(b);
	}

	private void drawFrame(Rectangle2D.Double b) {
		try {
			zoomGraphics.fill(b);
		} catch (Exception e) {
		}
	}

	/**
	 * Clears frame rectangle
	 * 
	 * @return zoomBounds rectangle
	 */
	public Rectangle2D.Double clearFrame() {
		Rectangle2D.Double r = new Rectangle2D.Double(zoomBounds.x,
				zoomBounds.y, zoomBounds.width, zoomBounds.height);
		try {
			drawFrame(zoomBounds);

			zoomGraphics.setPaintMode();
			zoomGraphics.dispose();
		} catch (Exception e) {
		}

		zoomBounds.setRect(0, 0, -1, -1);

		return r;
	}

	private JPopupMenu createPopup() {
		JMenuItem item;
		if (popup == null) {
			try {
				popup = new JPopupMenu();
				item = popup.add(new AxisFullScaleAction(this));
				item.setAccelerator(KeyStroke.getKeyStroke('1', 2));
				popup.add(new AxisZoomInAction(this));
				popup.add(new AxisZoomOutAction(this));
				popup.addSeparator();
				item = popup.add(new AxisPropertiesAction(this));
				item.setAccelerator(KeyStroke.getKeyStroke(
						java.awt.event.KeyEvent.VK_ENTER, 2));
			} catch (Exception e) {
			}
		}
		return popup;
	}

	// / --- interfaces

	/**
	 * @param evt
	 */
	public void mouseClicked(MouseEvent evt) {
	}

	/**
	 * @param evt
	 */
	public void mousePressed(MouseEvent evt) {
		if (evt.isPopupTrigger()) {
			createPopup().show(evt.getComponent(), evt.getX(), evt.getY());
			return;
		}

		zoomGraphics = (Graphics2D) getGraphics();

		if (evt.isControlDown()) {
			if (orientation == AXIS_HORIZONTAL) {
				zoomBounds.setRect(evt.getX(), getSize().height - tickLength,
						0, tickLength - 1);
			} else {
				zoomBounds.setRect(getSize().width - tickLength, evt.getY(),
						tickLength - 1, 0);
			}

			zoomGraphics.setXORMode(getBackground());
			zoomGraphics.setColor(scaleColor);
		}
	}

	/**
	 * @param evt
	 */
	public void mouseReleased(MouseEvent evt) {
		if (evt.isPopupTrigger()) {
			createPopup().show(evt.getComponent(), evt.getX(), evt.getY());
			return;
		}

		Rectangle2D.Double r = clearFrame();

		if (evt.isControlDown()) {
			if (evt.getClickCount() == 2) {
				this.fullScale();
			} else {
				try {
					if (orientation == AXIS_HORIZONTAL) {
						if (r.width > 1) {
							Point2D.Double a = new Point2D.Double();
							Point2D.Double b = new Point2D.Double();

							transform.inverseTransform(new Point2D.Double(r.x,
									r.y), a);
							b.x = r.width / transform.getScaleX();
							b.y = r.height / transform.getScaleY();

							setScale(new Rectangle2D.Double(a.x, a.y, b.x, b.y));
						}
					} else {
						if (r.height > 1) {
							Point2D.Double a = new Point2D.Double();
							Point2D.Double b = new Point2D.Double();

							transform.inverseTransform(new Point2D.Double(r.x,
									r.y), a);
							b.x = r.width / transform.getScaleX();
							b.y = r.height / transform.getScaleY();

							setScale(new Rectangle2D.Double(a.x, a.y, b.x, b.y));
						}
					}
				} catch (Exception e) {
				}
			}
		}
	}

	/**
	 * @param evt
	 */
	public void mouseDragged(MouseEvent evt) {
		if (evt.isControlDown()) {
			if (orientation == AXIS_HORIZONTAL) {
				flipFrame(new Rectangle2D.Double(zoomBounds.x, zoomBounds.y,
						evt.getX() - zoomBounds.x, zoomBounds.height));
			} else {
				flipFrame(new Rectangle2D.Double(zoomBounds.x, zoomBounds.y,
						zoomBounds.width, evt.getY() - zoomBounds.y));
			}
		}
	}

	/**
	 * @param evt
	 */
	public void mouseMoved(MouseEvent evt) {
	}

	/**
	 * @param evt
	 */
	public void mouseEntered(MouseEvent evt) {
	}

	/**
	 * @param evt
	 */
	public void mouseExited(MouseEvent evt) {
	}

	/**
	 * @param evt
	 */
	public void stateChanged(javax.swing.event.ChangeEvent evt) {
		if (orientation == AXIS_HORIZONTAL
				&& evt.getSource() == view.getScrollPane().getColumnHeader()) {
			changeLimits();
		}
		if (orientation == AXIS_VERTICAL
				&& evt.getSource() == view.getScrollPane().getRowHeader()) {
			changeLimits();
		}
	}
}