/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponces
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor;

import java.awt.Component;
import java.awt.Container;
import java.awt.Event;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;

import javax.swing.JDesktopPane;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;
import javax.swing.WindowConstants;

import ch.cern.lhcb.xmleditor.gui.TextView;
import ch.cern.lhcb.xmleditor.gui.TreeView;
import ch.cern.lhcb.xmleditor.specific.DtdFilesManager;
import ch.cern.lhcb.xmleditor.specific.IconsManager;
import ch.cern.lhcb.xmleditor.specific.Setup;
import ch.cern.lhcb.xmleditor.specific.XmlDisplaysManager;
import ch.cern.lhcb.xmleditor.specific.XmlFile;
import ch.cern.lhcb.xmleditor.specific.XmlFilesManager;
import ch.cern.lhcb.xmleditor.undo.UndoManager;

import com.simulogic.cese.UserHome;
import com.simulogic.utils.ExceptionDialog;

/**
 * The main class of the ch.cern.lhcb package. It handles the high level display
 * (multi windows, menu, ...)
 * 
 * @author Sebastien Ponce
 */
public class Editor extends JFrame implements ItemListener, ActionListener {

    /**
     * length of the shift between two window corners
     */
    private static final int FRAME_SHIFT = 20;

    /**
     * title of the main window
     */
    public static final String TITLE = ResourceManager
            .getString("Record_Editor"); //$NON-NLS-1$

    // Strings used for the different items of the menus
    private static final String FILE_MENU = ResourceManager.getString("File"); //$NON-NLS-1$
    private static final String OPEN_ITEM = ResourceManager.getString("Open"); //$NON-NLS-1$
    private static final String RECENT_ITEM = ResourceManager
            .getString("Recent"); //$NON-NLS-1$
    private static final String NEW_ITEM = ResourceManager.getString("New"); //$NON-NLS-1$
    private static final String GENERIC_NEW_ITEM = ResourceManager
            .getString("Generic"); //$NON-NLS-1$
    private static final String CLOSE_ITEM = ResourceManager.getString("Close"); //$NON-NLS-1$
    private static final String CLOSEALL_ITEM = ResourceManager
            .getString("Close_all"); //$NON-NLS-1$
    private static final String SAVE_ITEM = ResourceManager.getString("Save"); //$NON-NLS-1$
    private static final String SAVEAS_ITEM = ResourceManager
            .getString("Save_as"); //$NON-NLS-1$
    private static final String SAVEALL_ITEM = ResourceManager
            .getString("Save_all"); //$NON-NLS-1$
    private static final String EXIT_ITEM = ResourceManager.getString("Exit"); //$NON-NLS-1$
    private static final String EDIT_MENU = ResourceManager.getString("Edit"); //$NON-NLS-1$
    private static final String UNDO_ITEM = ResourceManager.getString("Undo"); //$NON-NLS-1$
    private static final String REDO_ITEM = ResourceManager.getString("Redo"); //$NON-NLS-1$
    private static final String UNDO_ALL_ITEM = ResourceManager
            .getString("Undo_all"); //$NON-NLS-1$
    private static final String REDO_ALL_ITEM = ResourceManager
            .getString("Redo_all"); //$NON-NLS-1$
    private static final String CUT_ITEM = ResourceManager.getString("Cut"); //$NON-NLS-1$
    private static final String COPY_ITEM = ResourceManager.getString("Copy"); //$NON-NLS-1$
    private static final String PASTE_ITEM = ResourceManager.getString("Paste"); //$NON-NLS-1$
    private static final String PASTE_INTO_PARENT_ITEM = ResourceManager
            .getString("Paste_into_Parent"); //$NON-NLS-1$
    private static final String OPTION_MENU = ResourceManager
            .getString("Option"); //$NON-NLS-1$
    private static final String RELOAD_SETUP_ITEM = ResourceManager
            .getString("Reload_setup"); //$NON-NLS-1$
    private static final String EDIT_SETUP_ITEM = ResourceManager
            .getString("Edit_setup"); //$NON-NLS-1$

    // Some of the menu items
    private JMenu recentMenuItem;
    private JMenu newMenuItem;
    private JMenuItem closeMenuItem;
    private JMenuItem closeAllMenuItem;
    private JMenuItem saveMenuItem;
    private JMenuItem saveAllMenuItem;
    private JMenuItem undoMenuItem;
    private JMenuItem undoAllMenuItem;
    private JMenuItem redoMenuItem;
    private JMenuItem redoAllMenuItem;

    /**
     * Number of frames already opened in this editor
     */
    private static HashMap openFramePositions = new HashMap();

    /**
     * the desktop manager
     */
    private JDesktopPane desktop;

    /**
     * a file manager that manages the different xml files used by the
     * application
     */
    private XmlFilesManager xmlFilesManager;

    /**
     * a file manager that manages the different dtd files used by the
     * application
     */
    private DtdFilesManager dtdFilesManager;

    /**
     * a manager that manages the different displaying policies of xml files
     */
    private XmlDisplaysManager xmlDisplaysManager;

    /**
     * an icon manager that manages the different icons used by the application
     */
    private IconsManager iconsManager;

    /**
     * an undo manager that manages undo and redo actions
     */
    private UndoManager undoManager;

    /**
     * the setup object for the editor
     */
    private Setup setup;

    /**
     * Constructor of this class
     * 
     * @param title the title of the main window
     */
    public Editor(String title) {
        // Super constructor
        super(title);
        // adds a window listener to catch the window closing
        addWindowListener(new XmlEditorListener());
        // creates a setup
        setup = new Setup(this);
        // sets up the GUI.
        desktop = new JDesktopPane();
        desktop.setPreferredSize(new java.awt.Dimension(640, 480));
        setContentPane(desktop);
        createMenuBar();
        // creates Managers common to all internal frames
        xmlFilesManager = new XmlFilesManager(this);
        dtdFilesManager = new DtdFilesManager(this);
        xmlDisplaysManager = new XmlDisplaysManager(this);
        iconsManager = new IconsManager(this);
        undoManager = new UndoManager(this);
        // initialize edit menu state
        this.updateEditMenu();
        this.pack();
    }

    /**
     * creates the menu bar for the main window
     */
    private void createMenuBar() {
        // creates an empty bar
        JMenuBar menuBar = new JMenuBar();

        // File Menu
        // //////////
        // creates the File menu, with mnemonic F
        JMenu fileMenu = new JMenu(FILE_MENU);
        fileMenu.setMnemonic(KeyEvent.VK_F);
        // creates some items and their mnemonics
        newMenuItem = new JMenu(NEW_ITEM);
        newMenuItem.setMnemonic(KeyEvent.VK_N);
        JMenuItem openMenuItem = new JMenuItem(OPEN_ITEM);
        openMenuItem.setMnemonic(KeyEvent.VK_O);
        openMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O,
                Event.CTRL_MASK));
        saveMenuItem = new JMenuItem(SAVE_ITEM);
        saveMenuItem.setMnemonic(KeyEvent.VK_S);
        saveMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S,
                Event.CTRL_MASK));
        saveAllMenuItem = new JMenuItem(SAVEALL_ITEM);
        saveAllMenuItem.setMnemonic(KeyEvent.VK_A);
        recentMenuItem = new JMenu(RECENT_ITEM);
        recentMenuItem.setMnemonic(KeyEvent.VK_R);
        closeMenuItem = new JMenuItem(CLOSE_ITEM);
        closeMenuItem.setMnemonic(KeyEvent.VK_C);
        closeMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W,
                Event.CTRL_MASK));
        closeAllMenuItem = new JMenuItem(CLOSEALL_ITEM);
        closeAllMenuItem.setMnemonic(KeyEvent.VK_L);
        JMenuItem exitMenuItem = new JMenuItem(EXIT_ITEM);
        exitMenuItem.setMnemonic(KeyEvent.VK_E);
        // adds a listener to each menuItem and to the menu itself
        openMenuItem.addActionListener(this);
        closeMenuItem.addActionListener(this);
        closeAllMenuItem.addActionListener(this);
        saveMenuItem.addActionListener(this);
        saveAllMenuItem.addActionListener(this);
        exitMenuItem.addActionListener(this);
        fileMenu.addItemListener(this);
        // builds the menu
        fileMenu.add(newMenuItem);
        fileMenu.add(openMenuItem);
        fileMenu.addSeparator();
        fileMenu.add(saveMenuItem);
        fileMenu.add(saveAllMenuItem);
        fileMenu.addSeparator();
        fileMenu.add(closeMenuItem);
        fileMenu.add(closeAllMenuItem);
        fileMenu.addSeparator();
        fileMenu.add(recentMenuItem);
        fileMenu.addSeparator();
        fileMenu.add(exitMenuItem);
        // if needed, adds the already open file names
        buildOpenFilesFileMenu();
        // if needed, adds the predefined data types
        buildDataTypesFileMenu();
        // adds the menu to the menu bar
        menuBar.add(fileMenu);

        // Edit Menu
        // //////////
        // creates the Edit menu, with mnemonic E
        JMenu editMenu = new JMenu(EDIT_MENU);
        editMenu.setMnemonic(KeyEvent.VK_E);
        // creates some items and their mnemonics
        undoMenuItem = new JMenuItem(UNDO_ITEM);
        undoMenuItem.setMnemonic(KeyEvent.VK_U);
        undoMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Z,
                Event.CTRL_MASK));
        undoAllMenuItem = new JMenuItem(UNDO_ALL_ITEM);
        undoAllMenuItem.setMnemonic(KeyEvent.VK_N);
        redoMenuItem = new JMenuItem(REDO_ITEM);
        redoMenuItem.setMnemonic(KeyEvent.VK_R);
        redoMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Y,
                Event.CTRL_MASK));
        redoAllMenuItem = new JMenuItem(REDO_ALL_ITEM);
        redoAllMenuItem.setMnemonic(KeyEvent.VK_E);
        JMenuItem cutMenuItem = new JMenuItem(CUT_ITEM);
        cutMenuItem.setMnemonic(KeyEvent.VK_C);
        cutMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X,
                Event.CTRL_MASK));
        JMenuItem copyMenuItem = new JMenuItem(COPY_ITEM);
        copyMenuItem.setMnemonic(KeyEvent.VK_O);
        copyMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C,
                Event.CTRL_MASK));
        JMenuItem pasteMenuItem = new JMenuItem(PASTE_ITEM);
        pasteMenuItem.setMnemonic(KeyEvent.VK_P);
        pasteMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V,
                Event.CTRL_MASK));
        JMenuItem pasteIntoParentMenuItem = new JMenuItem(
                PASTE_INTO_PARENT_ITEM);
        pasteIntoParentMenuItem.setMnemonic(KeyEvent.VK_Y);
        pasteIntoParentMenuItem.setAccelerator(KeyStroke.getKeyStroke(
                KeyEvent.VK_B, Event.CTRL_MASK));
        // adds a listener to each menuItem
        editMenu.addItemListener(this);
        undoMenuItem.addActionListener(this);
        undoAllMenuItem.addActionListener(this);
        redoMenuItem.addActionListener(this);
        redoAllMenuItem.addActionListener(this);
        cutMenuItem.addActionListener(this);
        copyMenuItem.addActionListener(this);
        pasteMenuItem.addActionListener(this);
        pasteIntoParentMenuItem.addActionListener(this);
        // builds the menu
        editMenu.add(undoMenuItem);
        editMenu.add(undoAllMenuItem);
        editMenu.add(redoMenuItem);
        editMenu.add(redoAllMenuItem);
        editMenu.addSeparator();
        editMenu.add(cutMenuItem);
        editMenu.add(copyMenuItem);
        editMenu.add(pasteMenuItem);
        editMenu.add(pasteIntoParentMenuItem);
        menuBar.add(editMenu);

        // Option Menu
        // ////////////
        // creates the Option menu, with mnemonic O
        JMenu optionMenu = new JMenu(OPTION_MENU);
        optionMenu.setMnemonic(KeyEvent.VK_O);
        // creates some items and their mnemonics
        JMenuItem reloadSetupMenuItem = new JMenuItem(RELOAD_SETUP_ITEM);
        reloadSetupMenuItem.setMnemonic(KeyEvent.VK_R);
        JMenuItem editSetupMenuItem = new JMenuItem(EDIT_SETUP_ITEM);
        editSetupMenuItem.setMnemonic(KeyEvent.VK_E);
        // adds a listener to each menuItem
        reloadSetupMenuItem.addActionListener(this);
        editSetupMenuItem.addActionListener(this);
        // builds the menu
        optionMenu.add(reloadSetupMenuItem);
        optionMenu.add(editSetupMenuItem);
        menuBar.add(optionMenu);

        // sets the menuBar
        setJMenuBar(menuBar);
    }

    /**
     * builds the part of the openFiles file menu, where already open files are.
     */
    private void buildOpenFilesFileMenu() {
        // adds files from Setup
        String[] files = setup.getLastOpenFiles();
        if (files != null) {
            if (files.length != 0) {
                for (int i = 0; i < files.length; i++) {
                    JMenuItem menuItem = new JMenuItem(files[i]);
                    menuItem.addActionListener(this);
                    recentMenuItem.add(menuItem);
                }
            }
        }
    }

    /**
     * builds the new file menu, where data types are.
     */
    private void buildDataTypesFileMenu() {
        // then adds types defined in the Setup
        Iterator types = setup.getDataTypes();
        if (types != null) {
            while (types.hasNext()) {
                JMenuItem menuItem = new JMenuItem((String) types.next());
                menuItem.addActionListener(this);
                newMenuItem.add(menuItem);
            }
        }
        newMenuItem.addSeparator();

        // adds the Generic type
        JMenuItem genericNewMenuItem = new JMenuItem(GENERIC_NEW_ITEM);
        genericNewMenuItem.setMnemonic(KeyEvent.VK_G);
        genericNewMenuItem.addActionListener(this);
        newMenuItem.add(genericNewMenuItem);

    }

    /**
     * rebuilds the openFiles file menu. It should be called when the list of
     * file already opened changed.
     */
    public void rebuildOpenFilesFileMenu() {
        // removes the old items
        recentMenuItem.removeAll();
        // adds the new ones
        buildOpenFilesFileMenu();
    }

    /**
     * rebuilds the new file menu
     */
    public void rebuildDataTypesFileMenu() {
        // removes the old items
        newMenuItem.removeAll();
        // adds the new ones
        buildDataTypesFileMenu();
    }

    /**
     * adds a new internal frame to this editor and selects it
     * 
     * @param frame the frame to be added
     */
    public void addInternalFrame(JInternalFrame frame) {
        // gets the first possible position for a frame
        int position = 0;
        while (openFramePositions.containsValue(new Integer(position))) {
            position++;
        }
        // updates the open frame positions
        openFramePositions.put(frame, new Integer(position));
        // sets the window's location.
        frame.setLocation(FRAME_SHIFT * position, FRAME_SHIFT * position);
        // adds the frame
        desktop.add(frame);
        // displays it
        frame.setVisible(true);
        try {
            frame.setSelected(true);
        } catch (Exception exception) {
        }
    }

    /**
     * called when an internalFrame is closing
     * 
     * @param frame the frame closing
     */
    public void internalFrameClosing(JInternalFrame frame) {
        openFramePositions.remove(frame);
    }

    /**
     * this creates a new TreeView inside the current Editor to display the file
     * named fileName
     * 
     * @param fileName the name of the new file to be displayed
     * @param putInOpenFiles says whether the open file should be put into the
     *        list of last open files or not
     */
    public void createNewTreeView(String fileName, boolean putInOpenFiles) {
        // creates the TreeView
        TreeView frame = new TreeView(fileName, this);
        // displays it
        this.addInternalFrame(frame);
        // adds the file to the list of open ones
        if (putInOpenFiles) {
            setup.addLastOpenFile(fileName);
        }
    }

    /**
     * this creates a new TreeView inside the current Editor
     * 
     * @param docType the name of the document type
     * @param dtd the file containing the dtd
     */
    public void createNewTreeView(String docType, File dtd) {
        // creates the TreeView
        TreeView frame = new TreeView(docType, dtd, this);
        // displays it
        this.addInternalFrame(frame);
    }

    /**
     * this creates a new TextView inside the current Editor to display the file
     * file
     * 
     * @param file the file to be displayed
     */
    public void createNewTextView(XmlFile file) {
        try {
            // gets the fileName
            String fileName;
            if (file.hasName) {
                fileName = file.getAbsolutePath();
            } else {
                fileName = file.getName();
            }
            TextView frame = new TextView(file.getContent(), fileName, this);
            this.addInternalFrame(frame);
        } catch (IOException exception) {
            exception.printStackTrace();
            ExceptionDialog
                    .showExceptionDialog(
                            "Error",
                            ResourceManager
                                    .getString("Unexpected_crash_during_getting_the_source.")
                                    + ResourceManager
                                            .getString("Please_report_this_error_to_the_LHCb_computing_team"),
                            exception);
        }
    }

    /**
     * checks that every file in this editor was saved. For those which are not,
     * it asks the user whether to save them.
     * 
     * @return if false, the current operation has to be canceled
     */
    public boolean checkSave() {
        // followsup has value 0 if all ok, -1 if the check should be stopped,
        // 1 if it should save every file without asking the user anymore and
        // -2 if the current operation should be canceled
        int followsup = 0;
        // just calls checkSave on every frame and stops if followsup is -1
        JInternalFrame[] frames = desktop.getAllFrames();
        for (int i = 0; i < frames.length; i++) {
            if (followsup >= 0) {
                if (frames[i] instanceof TreeView) {
                    followsup = ((TreeView) frames[i])
                            .checkSave(followsup == 0);
                    if (followsup == -2) {
                        // if -2 => cancel the current operation
                        return false;
                    }
                }
                // else it is view source windows. No problem with them
            } else {
                // -1 => just stop, don't cancel
                return true;
            }
        }
        return true;
    }

    /**
     * exits the application after making sure that all files have been saved or
     * have been canceled intentionnaly
     */
    private void exit() {
        // makes sure all open files were saved
        if (this.checkSave()) {
            // exit the application
            this.setVisible(false);
        } else {
            // cancels the closing
            this.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        }
    }

    /**
     * deals with action events occured on the main window
     * 
     * @param e an event
     */
    public void actionPerformed(ActionEvent e) {
        JMenuItem menuItem = (JMenuItem) e.getSource();
        String item = menuItem.getText();
        if (item.equals(OPEN_ITEM)) {
            // if the open menuItem was hit, activate the file chooser
            JFileChooser fileChooser = xmlFilesManager.getFileChooser();
            fileChooser.setDialogTitle(ResourceManager
                    .getString("Choose_the_file_to_open")); //$NON-NLS-1$
            int validFile = fileChooser.showOpenDialog(this);
            // check if the file was approved or canceled
            if (validFile == JFileChooser.APPROVE_OPTION) {
                // if a valid file was selected, open a new window
                String fileName = fileChooser.getSelectedFile()
                        .getAbsolutePath();
                this.createNewTreeView(fileName, true);
            }
            return;
        }
        if (item.equals(GENERIC_NEW_ITEM)) {
            // if the new menuItem was hit
            // activate a fileChooser to select a dtd for the new file
            JFileChooser fileChooser = xmlFilesManager.getFileChooser();
            fileChooser.setDialogTitle(ResourceManager
                    .getString("Select_a_dtd_file")); //$NON-NLS-1$
            int validFile = fileChooser.showDialog(this, ResourceManager
                    .getString("Use_as_dtd")); //$NON-NLS-1$
            // check if the file was approved or canceled
            if (validFile == JFileChooser.APPROVE_OPTION) {
                // if a valid file was selected, choose a document type
                String result = JOptionPane
                        .showInputDialog(
                                this,
                                ResourceManager
                                        .getString("Give_the_type_of_your_document"), ResourceManager.getString("DocType_Input"), //$NON-NLS-1$ //$NON-NLS-2$
                                JOptionPane.QUESTION_MESSAGE);
                // opens a new window, if needed
                if (result != null) {
                    this.createNewTreeView(result, fileChooser
                            .getSelectedFile());
                }
            }
            return;
        }
        if (item.equals(CLOSE_ITEM)) {
            // if close menuItem was hit, gets the frame concerned
            TreeView treeView = (TreeView) this.getSelectedFrame();
            // checks that all files were saved (or ask the user for it)
            // if the return value is -2, this means that the close operation
            // should be canceled
            if (!(treeView.checkSave(true) == -2)) {
                // closes the frame
                treeView.dispose();
            }
            return;
        }
        if (item.equals(CLOSEALL_ITEM)) {
            // checks that all files were saved (or ask the user for it)
            if (this.checkSave()) {
                // if closeAll menuItem was hit, close all frames
                JInternalFrame[] frames = desktop.getAllFrames();
                for (int i = 0; i < frames.length; i++) {
                    ((TreeView) frames[i]).dispose();
                }
            }
            return;
        }
        if (item.equals(SAVE_ITEM)) {
            // if save menuItem was hit, save the selected frame
            TreeView treeView = (TreeView) this.getSelectedFrame();
            treeView.save();
            return;
        }
        if (item.equals(SAVEALL_ITEM)) {
            // if saveAll menuItem was hit, save all frames
            JInternalFrame[] frames = desktop.getAllFrames();
            for (int i = 0; i < frames.length; i++) {
                ((TreeView) frames[i]).save();
            }
            return;
        }
        if (item.equals(EXIT_ITEM)) {
            // if exit menuItem was hit, exit
            this.exit();
            return;
        }
        if (menuItem.equals(undoMenuItem)) {
            undoManager.undo();
            return;
        }
        if (menuItem.equals(undoAllMenuItem)) {
            undoManager.undoAll();
            return;
        }
        if (menuItem.equals(redoMenuItem)) {
            undoManager.redo();
            return;
        }
        if (menuItem.equals(redoAllMenuItem)) {
            undoManager.redoAll();
            return;
        }
        if (item.equals(CUT_ITEM)) {
            // if cut menuItem was hit, informs the concerned tree
            TreeView treeView = (TreeView) this.getSelectedFrame();
            if (treeView != null) {
                treeView.getCutAndPaste().cut();
            }
            return;
        }
        if (item.equals(COPY_ITEM)) {
            // if copy menuItem was hit, informs the concerned tree
            TreeView treeView = (TreeView) this.getSelectedFrame();
            if (treeView != null) {
                treeView.getCutAndPaste().copy();
            }
            return;
        }
        if (item.equals(PASTE_ITEM)) {
            // if paste menuItem was hit, informs the concerned tree
            TreeView treeView = (TreeView) this.getSelectedFrame();
            if (treeView != null) {
                treeView.getCutAndPaste().paste();
            }
            return;
        }
        if (item.equals(PASTE_INTO_PARENT_ITEM)) {
            // if paste into parent menuItem was hit, informs the concerned tree
            TreeView treeView = (TreeView) this.getSelectedFrame();
            if (treeView != null) {
                treeView.getCutAndPaste().pasteIntoParent();
            }
            return;
        }
        if (item.equals(RELOAD_SETUP_ITEM)) {
            // if exit reloadSetupItem was hit, reloads the setup
            setup.reload();
            // updates the display of menus
            this.rebuildOpenFilesFileMenu();
            this.rebuildDataTypesFileMenu();
            // updates the IconsManager
            iconsManager.reload();
            // redisplay everything
            this.redisplayAll();
            return;
        }
        if (item.equals(EDIT_SETUP_ITEM)) {
            // if the editSetup menuItem was hit,
            // open the setup file for edition
            this.createNewTreeView(setup.getSetupFileName(), false);
            return;
        }
        // if we reach this point, then the menuItem that was hit is
        // either an already open file item or a data type
        Container popupMenu = ((JMenuItem) e.getSource()).getParent();
        if (popupMenu instanceof JPopupMenu) {
            Component invoker = ((JPopupMenu) popupMenu).getInvoker();
            if (invoker == recentMenuItem) {
                this.createNewTreeView(((JMenuItem) e.getSource()).getText(),
                        true);
            } else if (invoker == newMenuItem) {
                String type = ((JMenuItem) e.getSource()).getText();
                this.createNewTreeView(setup.getRootNodeFromType(type),
                        new File(UserHome.CESE_HOME + UserHome.FS
                                + setup.getDtdFromType(type)));
            }
        }
    }

    /**
     * deals with item events occured on the main window
     * 
     * @param e an event
     */
    public void itemStateChanged(ItemEvent e) {
        String item = ((JMenuItem) e.getSource()).getText();
        if (item.equals(FILE_MENU)) {
            // here we enable or disable the close and save items
            // depending of whether a frame is selected or not
            // and the closeAll and saveAll items
            // depending of whether there is an internal frame or not
            if (this.getSelectedFrame() != null) {
                closeMenuItem.setEnabled(true);
                saveMenuItem.setEnabled(true);
            } else {
                closeMenuItem.setEnabled(false);
                saveMenuItem.setEnabled(false);
            }
            if (desktop.getAllFrames().length > 0) {
                closeAllMenuItem.setEnabled(true);
                saveAllMenuItem.setEnabled(true);
            } else {
                closeAllMenuItem.setEnabled(false);
                saveAllMenuItem.setEnabled(false);
            }
        }
    }

    /**
     * updates the EditMenu depending on the state of the UndoManager
     */
    public void updateEditMenu() {
        if (undoManager.canUndo()) {
            undoMenuItem.setText(undoManager.getUndoPresentationName());
            undoMenuItem.setEnabled(true);
            undoAllMenuItem.setEnabled(true);
        } else {
            undoMenuItem.setText(UNDO_ITEM);
            undoMenuItem.setEnabled(false);
            undoAllMenuItem.setEnabled(false);
        }
        if (undoManager.canRedo()) {
            redoMenuItem.setText(undoManager.getRedoPresentationName());
            redoMenuItem.setEnabled(true);
            redoAllMenuItem.setEnabled(true);
        } else {
            redoMenuItem.setText(REDO_ITEM);
            redoMenuItem.setEnabled(false);
            redoAllMenuItem.setEnabled(false);
        }
    }

    /**
     * returns the selected frame among children, null if none
     * 
     * @return the selected Frame
     */
    private JInternalFrame getSelectedFrame() {
        // gets all the frames
        JInternalFrame[] frames = desktop.getAllFrames();
        if ((frames != null) && (frames.length > 0)) {
            // returns the first selected frame
            return frames[0];
        } else {
            // if no frame was selected, returns null
            return null;
        }
    }

    /**
     * redisplays every internal frame in this editor
     */
    public void redisplayAll() {
        JInternalFrame[] frames = desktop.getAllFrames();
        for (int i = 0; i < frames.length; i++) {
            ((TreeView) frames[i]).redisplayAll();
        }
    }

    /**
     * gets the XmlFilesManager associated to this ch.cern.lhcb
     * 
     * @return the XmlFilesManager
     */
    public XmlFilesManager getXmlFilesManager() {
        return xmlFilesManager;
    }

    /**
     * gets the DtdFilesManager associated to this ch.cern.lhcb
     * 
     * @return the DtdFilesManager
     */
    public DtdFilesManager getDtdFilesManager() {
        return dtdFilesManager;
    }

    /**
     * gets the XmlDisplaysManager associated to this ch.cern.lhcb
     * 
     * @return the XmlDisplaysManager
     */
    public XmlDisplaysManager getXmlDisplaysManager() {
        return xmlDisplaysManager;
    }

    /**
     * gets the IconsManager associated to this ch.cern.lhcb
     * 
     * @return the IconsManager
     */
    public IconsManager getIconsManager() {
        return iconsManager;
    }

    /**
     * gets the UndoManager associated to this ch.cern.lhcb
     * 
     * @return the UndoManager
     */
    public UndoManager getUndoManager() {
        return undoManager;
    }

    /**
     * gets the Setup associated to this editor
     * 
     * @return the Setup
     */
    public Setup getSetup() {
        return setup;
    }

    /**
     * this class only implements a windowListener in order to catch the closing
     * of the window
     */
    class XmlEditorListener extends WindowAdapter {

        /*
         * (non-Javadoc)
         * 
         * @see java.awt.event.WindowAdapter#windowClosing(java.awt.event.WindowEvent)
         */
        public void windowClosing(WindowEvent e) {
            Editor.this.exit();
        }
    }

}